import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.commonse.utilities as util
import wisdem.commonse.cylinder_member as member
from wisdem.commonse import gravity as g

NULL = member.NULL
NHEIGHT = 6
NPTS = member.get_nfull(NHEIGHT)
myones = np.ones((NPTS,))
secones = np.ones((NPTS - 1,))


class TestInputs(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # Test land based, 1 material
        self.inputs["s_in"] = np.linspace(0, 1, 5)
        self.inputs["layer_thickness"] = 0.25 * np.ones((1, 5))
        self.inputs["s_const1"] = 0.0
        self.inputs["s_const2"] = 0.0
        self.inputs["joint1"] = np.zeros(3)
        self.inputs["joint2"] = np.r_[np.zeros(2), 1e2]
        self.inputs["outer_diameter_in"] = 8 * np.ones(5)
        self.discrete_inputs["layer_materials"] = ["steel"]
        self.discrete_inputs["ballast_materials"] = ["slurry", "slurry", "seawater"]
        self.inputs["E_mat"] = 1e9 * np.ones((2, 3))
        self.inputs["E_user"] = 0.0
        self.inputs["G_mat"] = 1e8 * np.ones((2, 3))
        self.inputs["sigma_y_mat"] = np.array([1e7, 1e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.ones((2, 3))
        self.inputs["wohler_exp_mat"] = np.array([1e1, 1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1, 1e1])
        self.inputs["rho_mat"] = np.array([1e4, 1e5])
        self.inputs["rho_water"] = 1e3
        self.inputs["unit_cost_mat"] = np.array([1e1, 2e1])
        self.inputs["outfitting_factor_in"] = 1.05
        self.discrete_inputs["material_names"] = ["steel", "slurry"]
        self.opt = {}
        self.opt["n_height"] = [5]
        self.opt["n_layers"] = [1]
        self.opt["n_ballasts"] = [3]
        self.shape = "circular"

    def testDiscYAML_1Material(self):
        outputs = {}
        myobj = member.DiscretizationYAML(options=self.opt, idx=0, n_mat=2, shape=self.shape)
        myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        myones = np.ones(4)
        self.assertEqual(outputs["height"], 100.0)
        npt.assert_equal(outputs["section_height"], 25.0 * myones)
        npt.assert_equal(outputs["outer_diameter"], self.inputs["outer_diameter_in"])
        npt.assert_equal(outputs["wall_thickness"], 0.25 * myones)
        npt.assert_equal(outputs["E"], 1e9 * myones)
        npt.assert_equal(outputs["G"], 1e8 * myones)
        npt.assert_equal(outputs["sigma_y"], 1e7 * myones)
        npt.assert_equal(outputs["sigma_ult"], 1e7 * myones)
        npt.assert_equal(outputs["rho"], 1e4 * myones)
        npt.assert_equal(outputs["unit_cost"], 1e1 * myones)
        npt.assert_equal(outputs["outfitting_factor"], 1.05 * myones)
        npt.assert_equal(outputs["ballast_density"], np.array([1e5, 1e5, 1e3]))
        npt.assert_equal(outputs["ballast_unit_cost"], np.array([2e1, 2e1, 0.0]))
        A = np.pi * (16 - 3.75**2)
        I = (256.0 - 3.75**4) * np.pi / 4.0
        npt.assert_equal(outputs["z_param"], 100 * np.linspace(0, 1, 5))
        npt.assert_equal(outputs["sec_loc"], np.linspace(0, 1, 4))
        # npt.assert_equal(outputs["str_tw"], np.zeros(nout))
        # npt.assert_equal(outputs["tw_iner"], np.zeros(nout))
        npt.assert_equal(outputs["mass_den"], 1.05 * 1e4 * A * myones)
        npt.assert_equal(outputs["foreaft_iner"], 1.05 * 1e4 * I * myones)
        npt.assert_equal(outputs["sideside_iner"], 1.05 * 1e4 * I * myones)
        npt.assert_equal(outputs["foreaft_stff"], 1e9 * I * myones)
        npt.assert_equal(outputs["sideside_stff"], 1e9 * I * myones)
        npt.assert_equal(outputs["tor_stff"], 1e8 * 2 * I * myones)
        npt.assert_equal(outputs["axial_stff"], 1e9 * A * myones)
        # npt.assert_equal(outputs["cg_offst"], np.zeros(nout))
        # npt.assert_equal(outputs["sc_offst"], np.zeros(nout))
        # npt.assert_equal(outputs["tc_offst"], np.zeros(nout))

    def testDiscYAML_2Materials(self):
        outputs = {}

        # Test land based, 2 materials
        self.inputs["layer_thickness"] = np.array([[0.2, 0.2, 0.2, 0.0, 0.0], [0.0, 0.0, 0.0, 0.1, 0.1]])
        self.discrete_inputs["layer_materials"] = ["steel", "other"]
        self.inputs["E_mat"] = 1e9 * np.vstack((np.ones((2, 3)), 2 * np.ones((1, 3))))
        self.inputs["G_mat"] = 1e8 * np.vstack((np.ones((2, 3)), 2 * np.ones((1, 3))))
        self.inputs["sigma_y_mat"] = np.array([1e7, 1e7, 2e7])
        self.inputs["sigma_ult_mat"] = 1e7 * np.vstack((np.ones((2, 3)), 2 * np.ones((1, 3))))
        self.inputs["wohler_exp_mat"] = np.array([1e1, 1e1, 1e1])
        self.inputs["wohler_A_mat"] = np.array([1e1, 1e1, 1e1])
        self.inputs["rho_mat"] = np.array([1e4, 1e5, 2e4])
        self.inputs["unit_cost_mat"] = np.array([1e1, 2e1, 2e1])
        self.discrete_inputs["material_names"] = ["steel", "slurry", "other"]
        self.opt["n_layers"] = [2]
        myobj = member.DiscretizationYAML(options=self.opt, idx=0, n_mat=3, shape=self.shape)
        myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        # Define mixtures
        v = np.r_[np.mean([0.2, 0]), np.mean([0.1, 0.0])]
        vv = v / v.sum()
        x = np.r_[1, 2]
        xx1 = np.sum(x * vv)  # Mass weighted
        xx2 = 0.5 * np.sum(vv * x) + 0.5 / np.sum(vv / x)  # Volumetric method
        xx3 = np.sum(x * x * vv) / xx1  # Mass-cost weighted
        self.assertEqual(outputs["height"], 100.0)
        npt.assert_equal(outputs["section_height"], 25.0 * np.ones(4))
        npt.assert_equal(outputs["outer_diameter"], self.inputs["outer_diameter_in"])
        npt.assert_almost_equal(outputs["wall_thickness"], np.array([0.2, 0.2, v.sum(), 0.1]))
        npt.assert_almost_equal(outputs["E"], 1e9 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(outputs["G"], 1e8 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(outputs["sigma_y"], 1e7 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(outputs["sigma_ult"], 1e7 * np.array([1, 1, xx2, 2]))
        npt.assert_almost_equal(outputs["rho"], 1e4 * np.array([1, 1, xx1, 2]))
        npt.assert_almost_equal(outputs["unit_cost"], 1e1 * np.array([1, 1, xx3, 2]))
        npt.assert_equal(outputs["outfitting_factor"], 1.05 * np.ones(4))
        npt.assert_equal(outputs["ballast_density"], np.array([1e5, 1e5, 1e3]))
        npt.assert_equal(outputs["ballast_unit_cost"], np.array([2e1, 2e1, 0.0]))

    def test_sconst(self):
        outputs = {}
        self.inputs["s_const1"] = 0.1
        myobj = member.DiscretizationYAML(options=self.opt, idx=0, n_mat=2, shape=self.shape)
        myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(outputs["s"], np.array([0.0, 0.1, 0.5, 0.75, 1.0]))


class TestFullDiscretization(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}

        self.inputs["s"] = np.array([0.0, 0.1, 0.3, 0.6, 1.0])
        self.inputs["height"] = 1e1
        self.member_shape_variables = ["outer_diameter"]
        self.inputs["outer_diameter"] = 5.0 * np.ones(5)
        self.inputs["wall_thickness"] = 0.05 * np.ones(4)
        self.inputs["unit_cost"] = 1.0 * np.ones(4)
        self.inputs["E"] = 2e9 * np.ones(4)
        self.inputs["G"] = 2e7 * np.ones(4)
        self.inputs["sigma_y"] = 3e9 * np.ones(4)
        self.inputs["sigma_ult"] = 3e9 * np.ones(4)
        self.inputs["rho"] = 7850 * np.ones(4)
        self.inputs["outfitting_factor"] = 1.05 * np.ones(4)
        self.inputs["unit_cost"] = 7.0 * np.ones(4)
        self.inputs["joint1"] = np.zeros(3)
        self.inputs["joint2"] = np.r_[np.zeros(2), 1e2]

        self.mydis = member.MemberDiscretization(n_height=5, n_refine=2, member_shape_variables=self.member_shape_variables)

    def testRefine2(self):
        self.mydis.compute(self.inputs, self.outputs)
        npt.assert_array_equal(self.outputs["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.outputs["outer_diameter_full"], 5.0 * np.ones(9))
        npt.assert_array_equal(self.outputs["t_full"], 0.05 * np.ones(8))
        npt.assert_array_equal(self.outputs["E_full"], 2e9 * np.ones(8))
        npt.assert_array_equal(self.outputs["G_full"], 2e7 * np.ones(8))
        npt.assert_array_equal(self.outputs["nu_full"], 49 * np.ones(8))
        npt.assert_array_equal(self.outputs["sigma_y_full"], 3e9 * np.ones(8))
        # npt.assert_array_equal(self.outputs["sigma_ult_full"], 3e9 * np.ones(8))
        npt.assert_array_equal(self.outputs["rho_full"], 7850 * np.ones(8))
        npt.assert_array_equal(self.outputs["unit_cost_full"], 7 * np.ones(8))
        npt.assert_array_equal(self.outputs["outfitting_full"], 1.05 * np.ones(8))

        npt.assert_almost_equal(self.outputs["s_full"], np.array([0.0, 0.05, 0.1, 0.2, 0.3, 0.45, 0.6, 0.8, 1.0]))
        for k in self.inputs["s"]:
            self.assertIn(k, self.outputs["s_full"])

    def testRefineInterp(self):
        self.inputs["outer_diameter"] = np.array([5.0, 5.0, 6.0, 7.0, 7.0])
        self.inputs["wall_thickness"] = 1e-2 * np.array([5.0, 5.0, 6.0, 7.0])
        self.mydis.compute(self.inputs, self.outputs)
        npt.assert_almost_equal(self.outputs["s_full"], np.array([0.0, 0.05, 0.1, 0.2, 0.3, 0.45, 0.6, 0.8, 1.0]))
        npt.assert_array_equal(self.outputs["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.outputs["outer_diameter_full"], np.array([5.0, 5.0, 5.0, 5.5, 6.0, 6.5, 7.0, 7.0, 7.0]))
        npt.assert_array_equal(self.outputs["t_full"], 1e-2 * np.array([5.0, 5.0, 5.0, 5.0, 6.0, 6.0, 7.0, 7.0]))
        npt.assert_array_equal(self.outputs["E_full"], 2e9 * np.ones(8))
        npt.assert_array_equal(self.outputs["G_full"], 2e7 * np.ones(8))
        npt.assert_array_equal(self.outputs["nu_full"], 49 * np.ones(8))
        npt.assert_array_equal(self.outputs["sigma_y_full"], 3e9 * np.ones(8))
        # npt.assert_array_equal(self.outputs["sigma_ult_full"], 3e9 * np.ones(8))
        npt.assert_array_equal(self.outputs["rho_full"], 7850 * np.ones(8))
        npt.assert_array_equal(self.outputs["unit_cost_full"], 7 * np.ones(8))
        npt.assert_array_equal(self.outputs["outfitting_full"], 1.05 * np.ones(8))


class TestMemberComponent(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}

        self.inputs["s_full"] = np.linspace(0, 1, NPTS)
        self.inputs["z_full"] = 100 * np.linspace(0, 1, NPTS)
        self.inputs["height"] = 100
        self.inputs["outer_diameter_full"] = 10.0 * myones
        self.inputs["t_full"] = 0.05 * secones
        self.inputs["rho_full"] = 1e3 * secones
        self.inputs["E_full"] = 1e6 * secones
        self.inputs["G_full"] = 1e5 * secones
        self.inputs["sigma_y_full"] = 2e5 * secones
        # self.inputs["sigma_ult_full"] = 2e5 * secones
        self.inputs["outfitting_full"] = 1.1 * secones
        self.inputs["unit_cost_full"] = 1.0 * secones
        self.inputs["painting_cost_rate"] = 10.0
        self.inputs["labor_cost_rate"] = 2.0

        self.inputs["bulkhead_grid"] = np.array([0.0, 0.08, 0.16, 0.48, 0.88, 1.0])
        nbulk = len(self.inputs["bulkhead_grid"])
        self.inputs["bulkhead_thickness"] = 1.0 * np.ones(nbulk)

        self.inputs["ring_stiffener_web_thickness"] = 0.2
        self.inputs["ring_stiffener_flange_thickness"] = 0.3
        self.inputs["ring_stiffener_web_height"] = 0.5
        self.inputs["ring_stiffener_flange_width"] = 1.0
        self.inputs["ring_stiffener_spacing"] = 0.2  # non-dimensional ring stiffener spacing

        self.inputs["axial_stiffener_web_thickness"] = 0.0
        self.inputs["axial_stiffener_flange_thickness"] = 0.0
        self.inputs["axial_stiffener_web_height"] = 0.0
        self.inputs["axial_stiffener_flange_width"] = 0.0
        self.inputs["axial_stiffener_spacing"] = 0.0

        self.inputs["ballast_grid"] = np.array([[0.0, 0.08], [0.08, 0.16], [0.16, 0.48]])
        self.inputs["ballast_density"] = np.array([2e3, 4e3, 1e2])
        self.inputs["ballast_volume"] = np.pi * np.array([10.0, 10.0, 0.0])
        self.inputs["ballast_unit_cost"] = np.array([2.0, 4.0, 0.0])

        self.inputs["grid_axial_joints"] = np.array([0.44, 0.55, 0.66])
        self.inputs["joint1"] = np.array([20.0, 10.0, -30.0])
        self.inputs["joint2"] = np.array([25.0, 10.0, 15.0])
        self.inputs["s_ghost1"] = 0.0
        self.inputs["s_ghost2"] = 1.0

        self.inputs["total_mass_user"] = 0.0

        opt = {}
        opt["n_height"] = [NHEIGHT]
        opt["n_ballasts"] = [3]
        opt["n_bulkheads"] = [nbulk]
        opt["n_axial_joints"] = [3]
        opt["outer_shape"] = ["circular"]
        self.mem = member.MemberComplex(options=opt, idx=0)
        self.mem.setup()
        self.mem.sections = member.SortedDict()

    def testSortedDict(self):
        # Test create list
        self.mem.add_section(0.0, 0.5, "sec0")
        self.mem.add_section(0.5, 1.0, "sec1")
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.5, 1.0])
        self.assertEqual(list(self.mem.sections.values()), ["sec0", "sec1", None])

        # Test adding a node
        self.mem.add_node(0.25)
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.25, 0.5, 1.0])
        self.assertEqual(list(self.mem.sections.values()), ["sec0", "sec0", "sec1", None])
        self.mem.sections[0.25] = "sec2"
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.25, 0.5, 1.0])
        self.assertEqual(list(self.mem.sections.values()), ["sec0", "sec2", "sec1", None])
        self.mem.add_node(0.25)
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.25, 0.5, 1.0])
        self.assertEqual(list(self.mem.sections.values()), ["sec0", "sec2", "sec1", None])

        # Test inserting a section
        self.mem.insert_section(0.75, 0.8, "sec3")
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.25, 0.5, 0.75, 0.8, 1.0])
        self.assertEqual(list(self.mem.sections.values()), ["sec0", "sec2", "sec1", "sec3", "sec1", None])
        self.mem.insert_section(0.45, 0.55, "sec4")
        self.assertEqual(list(self.mem.sections.keys()), [0.0, 0.25, 0.45, 0.5, 0.55, 0.75, 0.8, 1.0])
        self.assertEqual(
            list(self.mem.sections.values()), ["sec0", "sec2", "sec4", "sec4", "sec1", "sec3", "sec1", None]
        )

    def testMainSections(self):
        self.mem.add_main_sections(self.inputs, self.outputs)

        m = np.pi * 0.25 * (10.0**2 - 9.9**2) * 1e3 * 1.1 * 100.0
        Iax = 0.5 * m * 0.25 * (10.0**2 + 9.9**2)
        Ix = (1 / 12.0) * m * (3 * 0.25 * (10.0**2 + 9.9**2) + 100**2) + m * 50 * 50  # parallel axis on last term
        self.assertAlmostEqual(self.outputs["shell_mass"], m)
        self.assertAlmostEqual(self.outputs["shell_z_cg"], 50.0)
        npt.assert_almost_equal(self.outputs["shell_I_base"], [Ix, Ix, Iax, 0.0, 0.0, 0.0], decimal=5)
        self.assertGreater(self.outputs["shell_cost"], 1e3)

        key = list(self.mem.sections.keys())
        self.assertEqual(key, self.inputs["s_full"].tolist())
        for k in key:
            if k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
            else:
                self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
                self.assertAlmostEqual(self.mem.sections[k].t, 0.05)
                self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
                self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
                self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)

    def testMainSectionsWithAxial(self):
        self.inputs["axial_stiffener_web_thickness"] = 0.2
        self.inputs["axial_stiffener_flange_thickness"] = 0.3
        self.inputs["axial_stiffener_web_height"] = 0.5
        self.inputs["axial_stiffener_flange_width"] = 1.0
        self.inputs["axial_stiffener_spacing"] = 0.25 * np.pi

        self.mem.add_main_sections(self.inputs, self.outputs)

        A_stiff = 0.2 * 0.5 + 1 * 0.3
        n_stiff = 8
        m = np.pi * 0.25 * (10.0**2 - 9.9**2) * 1e3 * 1.1 * 100.0
        Iax = 0.5 * m * 0.25 * (10.0**2 + 9.9**2)
        Ix = (1 / 12.0) * m * (3 * 0.25 * (10.0**2 + 9.9**2) + 100**2) + m * 50 * 50  # parallel axis on last term
        m += n_stiff * A_stiff * 1e3 * 1.1 * 100.0
        Iz_stiff = n_stiff * (0.2 * 0.5 * (0.5 * 9.9 - 0.25) ** 2 + 1 * 0.3 * (0.5 * 9.9 - 0.5 - 0.15) ** 2)
        Iax += Iz_stiff * 1e3 * 1.1 * 100
        self.assertAlmostEqual(self.outputs["shell_mass"], m)
        self.assertAlmostEqual(self.outputs["shell_z_cg"], 50.0)
        self.assertAlmostEqual(self.outputs["shell_I_base"][2], Iax, 5)
        self.assertGreater(self.outputs["shell_cost"], 1e3)

        key = list(self.mem.sections.keys())
        self.assertEqual(key, self.inputs["s_full"].tolist())
        for k in key:
            if k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
            else:
                self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
                self.assertAlmostEqual(self.mem.sections[k].t, 0.05 + n_stiff * A_stiff / (2 * np.pi * 4.7))
                self.assertAlmostEqual(
                    self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2) + n_stiff * A_stiff
                )
                self.assertAlmostEqual(
                    self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64 + Iz_stiff
                )
                self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
                self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
                self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)

    def testMainSectionsWithGhost(self):
        self.inputs["s_ghost1"] = 0.0
        self.inputs["s_ghost2"] = 0.9
        self.mem.add_main_sections(self.inputs, self.outputs)

        m = np.pi * 0.25 * (10.0**2 - 9.9**2) * 1e3 * 1.1 * 90.0
        Iax = 0.5 * m * 0.25 * (10.0**2 + 9.9**2)
        Ix = (1 / 12.0) * m * (3 * 0.25 * (10.0**2 + 9.9**2) + 90**2) + m * 45 * 45
        self.assertAlmostEqual(self.outputs["shell_mass"], m)
        self.assertAlmostEqual(self.outputs["shell_z_cg"], 45.0)
        npt.assert_almost_equal(self.outputs["shell_I_base"], [Ix, Ix, Iax, 0.0, 0.0, 0.0], decimal=5)
        self.assertGreater(self.outputs["shell_cost"], 0.9 * 1e3)
        key = list(self.mem.sections.keys())
        npt.assert_equal(key, np.unique(np.r_[0.9, self.inputs["s_full"].tolist()]))
        for k in key:
            if k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
            else:
                if k < 0.9:
                    self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.05)
                    self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                    self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)
                else:
                    self.assertAlmostEqual(self.mem.sections[k].D, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].A, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].J0, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e8)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e7)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e7)

    def testBulk(self):
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_bulkhead_sections(self.inputs, self.outputs)
        bgrid = self.inputs["bulkhead_grid"]

        s_full = self.inputs["s_full"]
        key = list(self.mem.sections.keys())
        bulks = np.vstack(([0.0, 0.01], np.c_[bgrid[1:-1] - 0.005, bgrid[1:-1] + 0.005], [0.99, 1.0]))
        expect = np.unique(np.r_[s_full, bulks.flatten()])
        npt.assert_almost_equal(key, expect)
        for k in key:
            inbulk = np.any(np.logical_and(k >= bulks[:, 0], k < bulks[:, 1]))
            if inbulk:
                self.assertAlmostEqual(self.mem.sections[k].t, 5.0)
                self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 0**2))
                self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
            elif k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
                continue
            else:
                self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].t, 0.05)
                self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)

            self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
            self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
            self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
            self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)

        nbulk = len(bgrid)
        R_i = 0.5 * 10 - 0.05
        m_bulk = 1.1 * 1e3 * np.pi * R_i**2 * 1.0
        npt.assert_almost_equal(self.outputs["bulkhead_mass"], m_bulk * nbulk)
        npt.assert_almost_equal(self.outputs["bulkhead_z_cg"], 100 * bgrid.mean())

        J0 = 0.50 * m_bulk * R_i**2
        I0 = 0.25 * m_bulk * R_i**2

        I = np.zeros(6)
        I[2] = nbulk * J0
        for k in bgrid:
            I[0] += I0 + m_bulk * (100 * k) ** 2
        I[1] = I[0]
        npt.assert_almost_equal(self.outputs["bulkhead_I_base"], I)

        self.assertGreater(self.outputs["bulkhead_cost"], 2e3)

    def testBulkWithGhost(self):
        self.inputs["s_ghost1"] = 0.0
        self.inputs["s_ghost2"] = 0.9
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_bulkhead_sections(self.inputs, self.outputs)
        bgrid = np.minimum(0.9, self.inputs["bulkhead_grid"])

        s_full = self.inputs["s_full"]
        key = list(self.mem.sections.keys())
        bulks = np.vstack(([0.0, 0.01], np.c_[bgrid[1:-1] - 0.005, bgrid[1:-1] + 0.005], [0.89, 0.9]))
        expect = np.unique(np.r_[s_full, 0.9, bulks.flatten()])
        npt.assert_almost_equal(key, expect)
        for k in key:
            inbulk = np.any(np.logical_and(k >= bulks[:, 0], k < bulks[:, 1]))
            if inbulk:
                self.assertAlmostEqual(self.mem.sections[k].t, 5.0)
                self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 0**2))
                self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 0**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
            elif k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
                continue
            else:
                if k < 0.9:
                    self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.05)
                    self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)
                else:
                    self.assertAlmostEqual(self.mem.sections[k].A, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].J0, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].D, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e8)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e7)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e7)

        nbulk = len(bgrid)
        R_i = 0.5 * 10 - 0.05
        m_bulk = 1.1 * 1e3 * np.pi * R_i**2 * 1.0
        npt.assert_almost_equal(self.outputs["bulkhead_mass"], m_bulk * nbulk)
        npt.assert_almost_equal(self.outputs["bulkhead_z_cg"], 100 * bgrid.mean())

        J0 = 0.50 * m_bulk * R_i**2
        I0 = 0.25 * m_bulk * R_i**2

        I = np.zeros(6)
        I[2] = nbulk * J0
        for k in bgrid:
            I[0] += I0 + m_bulk * (100 * k) ** 2
        I[1] = I[0]
        npt.assert_almost_equal(self.outputs["bulkhead_I_base"], I)

        self.assertGreater(self.outputs["bulkhead_cost"], 2e3)

    def testStiff(self):
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_ring_stiffener_sections(self.inputs, self.outputs)

        s_stiff = np.array([0.1, 0.3, 0.5, 0.7, 0.9])
        z_stiff = 100 * s_stiff

        Rwo = 0.5 * (10 - 2 * 0.05)
        Rwi = Rwo - 0.5
        Rfi = Rwi - 0.3
        self.assertEqual(self.outputs["flange_spacing_ratio"], 0.1)
        nout = np.where(self.outputs["stiffener_radius_ratio"] == NULL)[0][0]
        self.assertEqual(nout, 5)
        npt.assert_almost_equal(self.outputs["stiffener_radius_ratio"][nout:], NULL)
        npt.assert_almost_equal(self.outputs["stiffener_radius_ratio"][:nout], 1 - Rfi / 5)
        self.assertEqual(self.outputs["constr_flange_compactness"], 0.375 * np.sqrt(5) * 2 * 0.3)
        self.assertEqual(self.outputs["constr_web_compactness"], np.sqrt(5) * 0.2 / 0.5)

        # Test Mass
        A1 = np.pi * (Rwo**2 - Rwi**2)
        A2 = np.pi * (Rwi**2 - Rfi**2)
        V1 = A1 * 0.2
        V2 = A2 * 1.0
        m1 = V1 * 1e3
        m2 = V2 * 1e3
        m = m1 + m2
        f = 0.2
        self.assertAlmostEqual(self.outputs["stiffener_mass"], m * 5)
        self.assertAlmostEqual(self.outputs["stiffener_z_cg"], 50.0)
        self.assertGreater(self.outputs["stiffener_cost"], 1e3)

        # Test moment
        I_web = member.I_cyl(Rwi, Rwo, 0.2, m1)
        I_fl = member.I_cyl(Rfi, Rwi, 1.0, m2)
        I_sec = (I_web + I_fl).flatten()

        I = np.zeros(6)
        I[0] = np.sum(I_sec[0] + m * z_stiff**2.0)
        I[1] = I[0]
        I[2] = 5 * I_sec[2]
        npt.assert_almost_equal(self.outputs["stiffener_I_base"], I)

        s_full = self.inputs["s_full"]
        key = list(self.mem.sections.keys())
        stiffs = np.array([[0.095, 0.105], [0.295, 0.305], [0.495, 0.505], [0.695, 0.705], [0.895, 0.905]])
        expect = np.unique(np.r_[s_full, stiffs.flatten()])
        npt.assert_almost_equal(key, expect)
        print("HERE")
        for k in key:
            instiff = np.any(np.logical_and(k >= stiffs[:, 0], k < stiffs[:, 1]))
            if instiff:
                a = f * A1 + A2 + np.pi * 0.25 * (10.0**2 - 9.9**2)
                self.assertAlmostEqual(self.mem.sections[k].A, a)
                self.assertGreater(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertGreater(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertGreater(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].t, 5 - np.sqrt(25 - a / np.pi))
            elif k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
                continue
            else:
                self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].t, 0.05)

            self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
            self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
            self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
            self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
            self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)

    def testStiffWithGhost(self):
        self.inputs["s_ghost1"] = 0.0
        self.inputs["s_ghost2"] = 0.9
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_ring_stiffener_sections(self.inputs, self.outputs)

        s_stiff = np.array([0.1, 0.3, 0.5, 0.7])
        z_stiff = 100 * s_stiff

        Rwo = 0.5 * (10 - 2 * 0.05)
        Rwi = Rwo - 0.5
        Rfi = Rwi - 0.3
        self.assertEqual(self.outputs["flange_spacing_ratio"], 0.1)
        nout = np.where(self.outputs["stiffener_radius_ratio"] == NULL)[0][0]
        self.assertEqual(nout, 4)
        npt.assert_almost_equal(self.outputs["stiffener_radius_ratio"][nout:], NULL)
        npt.assert_almost_equal(self.outputs["stiffener_radius_ratio"][:nout], 1 - Rfi / 5)
        self.assertEqual(self.outputs["constr_flange_compactness"], 0.375 * np.sqrt(5) * 2 * 0.3)
        self.assertEqual(self.outputs["constr_web_compactness"], np.sqrt(5) * 0.2 / 0.5)

        # Test Mass
        A1 = np.pi * (Rwo**2 - Rwi**2)
        A2 = np.pi * (Rwi**2 - Rfi**2)
        V1 = A1 * 0.2
        V2 = A2 * 1.0
        m1 = V1 * 1e3
        m2 = V2 * 1e3
        m = m1 + m2
        f = 0.2
        self.assertAlmostEqual(self.outputs["stiffener_mass"], m * 4)
        self.assertAlmostEqual(self.outputs["stiffener_z_cg"], s_stiff.mean() * 100)
        self.assertGreater(self.outputs["stiffener_cost"], 1e3)

        # Test moment
        I_web = member.I_cyl(Rwi, Rwo, 0.2, m1)
        I_fl = member.I_cyl(Rfi, Rwi, 1.0, m2)
        I_sec = (I_web + I_fl).flatten()

        I = np.zeros(6)
        I[0] = np.sum(I_sec[0] + m * z_stiff**2.0)
        I[1] = I[0]
        I[2] = 4 * I_sec[2]
        npt.assert_almost_equal(self.outputs["stiffener_I_base"], I)

        s_full = self.inputs["s_full"]
        key = list(self.mem.sections.keys())
        stiffs = np.array([[0.095, 0.105], [0.295, 0.305], [0.495, 0.505], [0.695, 0.705]])
        expect = np.unique(np.r_[s_full, 0.9, stiffs.flatten()])
        npt.assert_almost_equal(key, expect)
        for k in key:
            instiff = np.any(np.logical_and(k >= stiffs[:, 0], k < stiffs[:, 1]))
            if instiff:
                a = f * A1 + A2 + np.pi * 0.25 * (10.0**2 - 9.9**2)
                self.assertAlmostEqual(self.mem.sections[k].A, a)
                self.assertGreater(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertGreater(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertGreater(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                self.assertAlmostEqual(self.mem.sections[k].t, 5 - np.sqrt(25 - a / np.pi))
                self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
            elif k == 1.0:
                self.assertEqual(self.mem.sections[k], None)
                continue
            else:
                if k < 0.9:
                    self.assertAlmostEqual(self.mem.sections[k].A, np.pi * 0.25 * (10.0**2 - 9.9**2))
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].J0, 2 * np.pi * (10.0**4 - 9.9**4) / 64)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.05)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 1.1 * 1e3)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e6)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e5)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e5)
                    self.assertAlmostEqual(self.mem.sections[k].D, 10.0)
                else:
                    self.assertAlmostEqual(self.mem.sections[k].A, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Ixx, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].Iyy, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].J0, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].t, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].rho, 0.01)
                    self.assertAlmostEqual(self.mem.sections[k].E, 1e8)
                    self.assertAlmostEqual(self.mem.sections[k].G, 1e7)
                    self.assertAlmostEqual(self.mem.sections[k].sigy, 2e7)
                    self.assertAlmostEqual(self.mem.sections[k].D, 0.01)

    def testBallast(self):
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_circular_ballast_sections(self.inputs, self.outputs)

        area = 0.25 * np.pi * 9.9**2
        h = 10 * np.pi / area
        cg_perm = (2 * 0.5 * h + 4 * (8 + 0.5 * h)) / 6
        m_perm = np.pi * 6e4

        I_perm = np.zeros(6)
        I_perm[2] = 0.5 * m_perm * 0.25 * 9.9**2
        I_perm[0] = (
            m_perm * (3 * 0.25 * 9.9**2 + h**2) / 12.0
            + (1 / 3) * m_perm * (0.5 * h) ** 2
            + (2 / 3) * m_perm * (8 + 0.5 * h) ** 2
        )
        I_perm[1] = I_perm[0]

        self.assertAlmostEqual(self.outputs["ballast_mass"], m_perm)
        self.assertAlmostEqual(self.outputs["ballast_cost"], np.pi * 20e4)
        self.assertAlmostEqual(self.outputs["ballast_z_cg"], cg_perm)
        npt.assert_almost_equal(self.outputs["ballast_height"], np.r_[h, h, 0.0] / 100)
        npt.assert_almost_equal(self.outputs["ballast_I_base"], I_perm)
        self.assertAlmostEqual(self.outputs["variable_ballast_capacity"], area * 32)
        npt.assert_almost_equal(self.outputs["variable_ballast_Vpts"], area * 32 / 9.0 * np.arange(10))
        npt.assert_almost_equal(self.outputs["variable_ballast_spts"], np.linspace(0.16, 0.48, 10))

    def testBallastWithGhost(self):
        self.inputs["s_ghost1"] = 0.1
        self.inputs["s_ghost2"] = 1.0
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_circular_ballast_sections(self.inputs, self.outputs)

        area = 0.25 * np.pi * 9.9**2
        h = 10 * np.pi / area
        cg_perm = (2 * (10 + 0.5 * h) + 4 * (18 + 0.5 * h)) / 6
        m_perm = np.pi * 6e4

        I_perm = np.zeros(6)
        I_perm[2] = 0.5 * m_perm * 0.25 * 9.9**2
        I_perm[0] = (
            m_perm * (3 * 0.25 * 9.9**2 + h**2) / 12.0
            + (1 / 3) * m_perm * (10 + 0.5 * h) ** 2
            + (2 / 3) * m_perm * (18 + 0.5 * h) ** 2
        )
        I_perm[1] = I_perm[0]

        self.assertAlmostEqual(self.outputs["ballast_mass"], m_perm)
        self.assertAlmostEqual(self.outputs["ballast_cost"], np.pi * 20e4)
        self.assertAlmostEqual(self.outputs["ballast_z_cg"], cg_perm)
        npt.assert_almost_equal(self.outputs["ballast_I_base"], I_perm, 6)
        self.assertAlmostEqual(self.outputs["variable_ballast_capacity"], area * 32)
        npt.assert_almost_equal(self.outputs["variable_ballast_Vpts"], area * 32 / 9.0 * np.arange(10))
        npt.assert_almost_equal(self.outputs["variable_ballast_spts"], np.linspace(0.26, 0.58, 10))

    def testMassProp(self):
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_bulkhead_sections(self.inputs, self.outputs)
        self.mem.add_ring_stiffener_sections(self.inputs, self.outputs)
        self.mem.add_circular_ballast_sections(self.inputs, self.outputs)
        self.mem.compute_mass_properties(self.inputs, self.outputs)

        m_shell = np.pi * 0.25 * (10.0**2 - 9.9**2) * 1e3 * 1.1 * 100.0
        R_i = 0.5 * 10 - 0.05
        cg_shell = 50

        nbulk = len(self.inputs["bulkhead_grid"])
        m_bulk = 1.1 * 1e3 * np.pi * R_i**2 * 1.0
        cg_bulk = 100 * self.inputs["bulkhead_grid"].mean()

        Rwo = 0.5 * (10 - 2 * 0.05)
        Rwi = Rwo - 0.5
        Rfi = Rwi - 0.3
        A1 = np.pi * (Rwo**2 - Rwi**2)
        A2 = np.pi * (Rwi**2 - Rfi**2)
        V1 = A1 * 0.2
        V2 = A2 * 1.0
        m1 = V1 * 1e3
        m2 = V2 * 1e3
        m_stiff = m1 + m2
        cg_stiff = 50.0

        area = 0.25 * np.pi * 9.9**2
        h = 10 * np.pi / area
        cg_perm = (2 * 0.5 * h + 4 * (8 + 0.5 * h)) / 6
        m_perm = np.pi * 6e4

        m_tot = m_shell + nbulk * m_bulk + 5 * m_stiff + m_perm
        self.assertAlmostEqual(self.outputs["total_mass"], m_tot)
        self.assertAlmostEqual(self.outputs["structural_mass"], m_tot - m_perm)
        self.assertAlmostEqual(
            self.outputs["z_cg"], (50 * (m_shell + 5 * m_stiff) + nbulk * m_bulk * cg_bulk + m_perm * cg_perm) / m_tot
        )
        self.assertEqual(
            self.outputs["total_cost"],
            self.outputs["shell_cost"]
            + self.outputs["ballast_cost"]
            + self.outputs["bulkhead_cost"]
            + self.outputs["stiffener_cost"],
        )
        self.assertEqual(
            self.outputs["structural_cost"],
            self.outputs["shell_cost"] + self.outputs["bulkhead_cost"] + self.outputs["stiffener_cost"],
        )

    def testNodalFinish(self):
        self.outputs["z_cg"] = 40.0
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.nodal_discretization(self.inputs, self.outputs)

        s_full = self.inputs["s_full"]
        s_all = self.outputs["s_all"]
        nout = np.where(s_all == NULL)[0][0]
        self.assertEqual(nout, len(s_full) + 3)
        npt.assert_almost_equal(s_all[nout:], NULL)
        npt.assert_almost_equal(s_all[:nout], np.sort(np.r_[s_full, 0.44, 0.55, 0.66]))

        npt.assert_almost_equal(self.outputs["center_of_mass"], np.array([22, 10, -12]))
        npt.assert_almost_equal(self.outputs["nodes_xyz_all"][nout:, :], NULL)
        npt.assert_almost_equal(self.outputs["nodes_xyz_all"][:nout, 0], 20 + s_all[:nout] * 5)
        npt.assert_almost_equal(self.outputs["nodes_xyz_all"][:nout, 1], 10)
        npt.assert_almost_equal(self.outputs["nodes_xyz_all"][:nout, 2], -30 + s_all[:nout] * 45)

        nelem = nout - 1
        for var in ["D", "t", "A", "Ixx", "Iyy", "J0", "rho", "G", "E", "sigma_y"]:
            npt.assert_almost_equal(self.outputs["section_" + var][nelem:], NULL)
        npt.assert_almost_equal(self.outputs["section_D"][:nelem], 10.0)
        npt.assert_almost_equal(self.outputs["section_t"][:nelem], 0.05)
        npt.assert_almost_equal(self.outputs["section_Ixx"][:nelem], np.pi * (10.0**4 - 9.9**4) / 64)
        npt.assert_almost_equal(self.outputs["section_Iyy"][:nelem], np.pi * (10.0**4 - 9.9**4) / 64)
        npt.assert_almost_equal(self.outputs["section_J0"][:nelem], 2 * np.pi * (10.0**4 - 9.9**4) / 64)
        npt.assert_almost_equal(self.outputs["section_rho"][:nelem], 1.1 * 1e3)
        npt.assert_almost_equal(self.outputs["section_E"][:nelem], 1e6)
        npt.assert_almost_equal(self.outputs["section_G"][:nelem], 1e5)
        npt.assert_almost_equal(self.outputs["section_sigma_y"][:nelem], 2e5)
        # npt.assert_almost_equal(self.outputs["section_sigma_ult"][:nelem], 2e5)

    def testCompute(self):
        self.mem.compute(self.inputs, self.outputs)
        # 2 points added for bulkheads and stiffeners
        # Bulkheads at 0,1 only get 1 new point
        nbulk = len(self.inputs["bulkhead_grid"])
        s_all = self.outputs["s_all"]
        nout = np.where(s_all == NULL)[0][0]
        self.assertEqual(nout, NPTS + 3 + 2 * nbulk - 2 + 2 * 5)

    def testUserMass(self):
        self.inputs["total_mass_user"] = 123456.0
        self.mem.compute(self.inputs, self.outputs)
        self.assertEqual(self.outputs["total_mass"], 123456.0)

    def testDeconflict(self):
        self.inputs["bulkhead_grid"] = np.array([0.0, 0.1, 1.0])
        self.mem.add_main_sections(self.inputs, self.outputs)
        self.mem.add_ring_stiffener_sections(self.inputs, self.outputs)

        s_full = self.inputs["s_full"]
        key = list(self.mem.sections.keys())
        stiffs = np.array([[0.075, 0.085], [0.295, 0.305], [0.495, 0.505], [0.695, 0.705], [0.895, 0.905]])
        expect = np.unique(np.r_[s_full, stiffs.flatten()])
        npt.assert_almost_equal(key, expect)


class TestHydro(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        # For Geometry call
        n_height = 4
        npts = member.get_nfull(n_height)
        self.inputs["s_full"] = np.linspace(0, 1.0, npts)
        self.inputs["z_full"] = np.linspace(0, 50.0, npts)
        self.inputs["outer_diameter_full"] = 10.0 * np.ones(npts)
        self.inputs["s_all"] = np.linspace(0, 1.0, 2 * npts)
        self.inputs["nodes_xyz"] = np.c_[
            1 * np.ones(2 * npts), 2 * np.ones(2 * npts), np.linspace(0, 50.0, 2 * npts) - 75
        ]
        self.inputs["rho_water"] = 1e3
        self.inputs["s_ghost1"] = 0.0
        self.inputs["s_ghost2"] = 1.0
        self.inputs["joint1"] = self.inputs["nodes_xyz"][0, :]
        self.inputs["joint2"] = self.inputs["nodes_xyz"][-1, :]

        self.hydro = member.MemberHydro(n_full=npts)

    def testVerticalSubmerged(self):
        npts = self.inputs["s_full"].size
        self.hydro.compute(self.inputs, self.outputs)

        rho_w = self.inputs["rho_water"]
        V_expect = np.pi * 25.0 * 50.0
        cb_expect = np.array([1.0, 2.0, -50])
        Ixx = 0  # 0.25 * np.pi * 1e4
        Axx = 0  # np.pi * 1e2
        self.assertAlmostEqual(self.outputs["displacement"], V_expect)
        self.assertAlmostEqual(self.outputs["buoyancy_force"], V_expect * rho_w * g)
        npt.assert_almost_equal(self.outputs["center_of_buoyancy"], cb_expect)
        self.assertEqual(self.outputs["idx_cb"], npts - 1)  # Halfway node point
        self.assertAlmostEqual(self.outputs["Iwaterx"], Ixx)
        self.assertAlmostEqual(self.outputs["Awater"], Axx)
        npt.assert_equal(self.outputs["waterline_centroid"], [0.0, 0.0])
        npt.assert_almost_equal(self.outputs["z_dim"], np.linspace(0, 50.0, npts) - 75)
        npt.assert_almost_equal(self.outputs["d_eff"], self.inputs["outer_diameter_full"])

        m_a = np.zeros(6)
        m_a[:2] = V_expect * rho_w
        m_a[2] = 0.5 * (8.0 / 3.0) * rho_w * 125
        m_a[3:5] = np.pi * rho_w * 25.0 * ((-25 - cb_expect[-1]) ** 3.0 - (-75 - cb_expect[-1]) ** 3.0) / 3.0
        npt.assert_almost_equal(self.outputs["added_mass"], m_a, decimal=-5)

        self.inputs["s_ghost1"] = 0.25
        self.inputs["s_ghost2"] = 0.75
        self.hydro.compute(self.inputs, self.outputs)
        self.assertAlmostEqual(self.outputs["displacement"], 0.5 * V_expect)

    def testVerticalWaterplane(self):
        npts = self.inputs["s_full"].size
        self.inputs["nodes_xyz"] = np.c_[1 * np.ones(npts), 2 * np.ones(npts), np.linspace(0, 50.0, npts) - 25]
        self.inputs["joint1"] = self.inputs["nodes_xyz"][0, :]
        self.inputs["joint2"] = self.inputs["nodes_xyz"][-1, :]
        self.hydro.compute(self.inputs, self.outputs)

        rho_w = self.inputs["rho_water"]
        V_expect = np.pi * 25.0 * 25.0
        cb_expect = np.array([1.0, 2.0, -12.5])
        Ixx = 0.25 * np.pi * 625
        Axx = np.pi * 25
        self.assertAlmostEqual(self.outputs["displacement"], V_expect)
        self.assertAlmostEqual(self.outputs["buoyancy_force"], V_expect * rho_w * g)
        npt.assert_almost_equal(self.outputs["center_of_buoyancy"], cb_expect)
        self.assertEqual(self.outputs["idx_cb"], int(0.5 * npts - 1))
        self.assertAlmostEqual(self.outputs["Iwaterx"], Ixx)
        self.assertAlmostEqual(self.outputs["Awater"], Axx)
        npt.assert_equal(self.outputs["waterline_centroid"], [1.0, 2.0])
        npt.assert_almost_equal(self.outputs["z_dim"], np.linspace(0, 50.0, npts) - 25)
        npt.assert_almost_equal(self.outputs["d_eff"], self.inputs["outer_diameter_full"])

        m_a = np.zeros(6)
        m_a[:2] = V_expect * rho_w
        m_a[2] = 0.5 * (8.0 / 3.0) * rho_w * 125
        m_a[3:5] = np.pi * rho_w * 25.0 * ((0 - cb_expect[-1]) ** 3.0 - (-25 - cb_expect[-1]) ** 3.0) / 3.0
        npt.assert_almost_equal(self.outputs["added_mass"], m_a, decimal=-5)

    def test45deg(self):
        npts = self.inputs["s_full"].size
        xy = np.linspace(0, 50.0, npts) - 25
        self.inputs["nodes_xyz"] = np.c_[xy, np.zeros(npts), xy]
        self.inputs["joint1"] = self.inputs["nodes_xyz"][0, :]
        self.inputs["joint2"] = self.inputs["nodes_xyz"][-1, :]
        self.hydro.compute(self.inputs, self.outputs)

        rho_w = self.inputs["rho_water"]
        V_expect = np.pi * 25.0 * 25.0
        cb_expect = np.array([-12.5, 0.0, -12.5])
        Ixx = 0.25 * np.pi * 625
        Axx = np.pi * 25
        self.assertAlmostEqual(self.outputs["displacement"], V_expect)
        self.assertAlmostEqual(self.outputs["buoyancy_force"], V_expect * rho_w * g)
        npt.assert_almost_equal(self.outputs["center_of_buoyancy"], cb_expect)
        self.assertEqual(self.outputs["idx_cb"], int(0.5 * npts - 1))
        self.assertAlmostEqual(self.outputs["Iwaterx"], Ixx)
        self.assertAlmostEqual(self.outputs["Awater"], Axx)
        npt.assert_almost_equal(self.outputs["waterline_centroid"], [0.0, 0.0])
        npt.assert_almost_equal(self.outputs["z_dim"], np.linspace(0, 50.0, npts) - 25)
        npt.assert_almost_equal(self.outputs["d_eff"], self.inputs["outer_diameter_full"] / np.cos(0.25 * np.pi))

        m_a = np.zeros(6)
        m_a[:2] = V_expect * rho_w
        m_a[2] = 0.5 * (8.0 / 3.0) * rho_w * 125
        m_a[3:5] = np.pi * rho_w * 25.0 * ((0 - cb_expect[-1]) ** 3.0 - (-25 - cb_expect[-1]) ** 3.0) / 3.0
        for k in range(6):
            self.assertAlmostEqual(self.outputs["added_mass"][k], m_a[k], -5)


class TestGlobal2Member(unittest.TestCase):
    def testMemax(self):
        n_height = 4
        npts = member.get_nfull(n_height)
        myobj = member.Global2MemberLoads(n_full=npts, memmax=True)

        inputs = {}
        outputs = {}
        inputs["s_full"] = np.linspace(0, 1.0, npts)
        inputs["s_all"] = NULL * np.ones(member.MEMMAX)
        inputs["s_all"][: 2 * npts] = np.linspace(0, 1.0, 2 * npts)
        inputs["joint1"] = np.zeros(3)
        inputs["joint2"] = np.r_[np.zeros(2), 1e2]

        Px = NULL * np.ones(member.MEMMAX)
        Py = NULL * np.ones(member.MEMMAX)
        Pz = NULL * np.ones(member.MEMMAX)
        qdyn = NULL * np.ones(member.MEMMAX)
        inputs["Px_global"] = np.zeros(npts)
        inputs["Py_global"] = np.zeros(npts)
        inputs["Pz_global"] = np.ones(npts)
        inputs["qdyn_global"] = 2 * np.ones(npts)
        Px[: (2 * npts)] = 1.0
        Py[: (2 * npts)] = 0.0
        Pz[: (2 * npts)] = 0.0
        qdyn[: (2 * npts)] = 2.0

        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px"], Px)
        npt.assert_almost_equal(outputs["Py"], Py)
        npt.assert_almost_equal(outputs["Pz"], Pz)
        npt.assert_almost_equal(outputs["qdyn"], qdyn)

        inputs["Px_global"] = np.ones(npts)
        inputs["Py_global"] = np.zeros(npts)
        inputs["Pz_global"] = np.zeros(npts)
        Px[: (2 * npts)] = 0.0
        Py[: (2 * npts)] = 0.0
        Pz[: (2 * npts)] = -1.0
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px"], Px)
        npt.assert_almost_equal(outputs["Py"], Py)
        npt.assert_almost_equal(outputs["Pz"], Pz)
        npt.assert_almost_equal(outputs["qdyn"], qdyn)

    def testFull(self):
        n_height = 4
        npts = member.get_nfull(n_height)
        myobj = member.Global2MemberLoads(n_full=npts, memmax=False)

        inputs = {}
        outputs = {}
        inputs["s_full"] = np.linspace(0, 1.0, npts)
        inputs["s_all"] = np.linspace(0, 1.0, 2 * npts)
        inputs["joint1"] = np.zeros(3)
        inputs["joint2"] = np.r_[np.zeros(2), 1e2]

        inputs["Px_global"] = np.zeros(npts)
        inputs["Py_global"] = np.zeros(npts)
        inputs["Pz_global"] = np.ones(npts)
        inputs["qdyn_global"] = 2 * np.ones(npts)
        Px = np.ones(npts)
        Py = np.zeros(npts)
        Pz = np.zeros(npts)
        qdyn = 2 * np.ones(npts)

        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px"], Px)
        npt.assert_almost_equal(outputs["Py"], Py)
        npt.assert_almost_equal(outputs["Pz"], Pz)
        npt.assert_almost_equal(outputs["qdyn"], qdyn)

        inputs["Px_global"] = np.ones(npts)
        inputs["Py_global"] = np.zeros(npts)
        inputs["Pz_global"] = np.zeros(npts)
        Px = np.zeros(npts)
        Py = np.zeros(npts)
        Pz = -np.ones(npts)
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["Px"], Px)
        npt.assert_almost_equal(outputs["Py"], Py)
        npt.assert_almost_equal(outputs["Pz"], Pz)
        npt.assert_almost_equal(outputs["qdyn"], qdyn)


class TestGroup(unittest.TestCase):
    def testStandard(self):
        opt = {}
        opt["n_height"] = [5]
        opt["n_layers"] = [1]
        opt["n_bulkheads"] = nbulk = [0]
        opt["n_ballasts"] = [0]
        opt["n_axial_joints"] = [0]

        prob = om.Problem(reports=False)

        prob.model = member.MemberStandard(column_options=opt, idx=0, n_mat=2)

        prob.setup()
        prob["s_in"] = np.linspace(0, 1, 5)
        prob["layer_thickness"] = 0.05 * np.ones((1, 5))
        prob["height"] = 1e2
        prob["outer_diameter_in"] = 10 * np.ones(5)
        prob["layer_materials"] = ["steel"]
        prob["ballast_materials"] = ["slurry", "seawater"]
        prob["E_mat"] = 1e9 * np.ones((2, 3))
        prob["G_mat"] = 1e8 * np.ones((2, 3))
        prob["sigma_y_mat"] = np.array([1e7, 1e7])
        prob["sigma_ult_mat"] = 1e7 * np.ones((2, 3))
        prob["wohler_exp_mat"] = np.array([1e1, 1e1])
        prob["wohler_A_mat"] = np.array([1e1, 1e1])
        prob["rho_mat"] = np.array([1e4, 1e5])
        prob["rho_water"] = 1025.0
        prob["unit_cost_mat"] = np.array([1e1, 2e1])
        prob["outfitting_factor_in"] = 1.1
        prob["material_names"] = ["steel", "slurry"]
        prob["painting_cost_rate"] = 10.0
        prob["labor_cost_rate"] = 2.0

        prob["joint1"] = np.array([20.0, 10.0, -30.0])
        prob["joint2"] = np.array([25.0, 10.0, 15.0])
        prob["s_ghost1"] = 0.0
        prob["s_ghost2"] = 1.0

        prob.run_model()
        out_list = prob.model.list_outputs(prom_name=True, units=False, out_stream=None)
        for k in out_list:
            if np.all(k[1]["val"] == 0.0) or np.all(k[1]["val"] == NULL):
                name = k[1]["prom_name"]
                if (
                    name.find("Py") > 0
                    or name.find("Pz") > 0
                    or name.find("beta") > 0
                    or name.find("offst") > 0
                    or name.find("tw") >= 0
                    or name.find("ballast") >= 0
                ):
                    continue
                print(f"{name} is all zero!")
        self.assertTrue(True)

    def testDetailed(self):
        opt = {}
        opt["n_height"] = [5]
        opt["n_layers"] = [1]
        opt["n_bulkheads"] = nbulk = [4]
        opt["n_ballasts"] = [2]
        opt["n_axial_joints"] = [3]
        opt["outer_shape"] = ["circular"]

        prob = om.Problem(reports=False)

        prob.model = member.MemberDetailed(column_options=opt, idx=0, n_mat=2)

        prob.setup()
        prob["s_in"] = np.linspace(0, 1, 5)
        prob["layer_thickness"] = 0.05 * np.ones((1, 5))
        prob["height"] = 1e2
        prob["outer_diameter_in"] = 10 * np.ones(5)
        prob["layer_materials"] = ["steel"]
        prob["ballast_materials"] = ["slurry", "seawater"]
        prob["E_mat"] = 1e9 * np.ones((2, 3))
        prob["G_mat"] = 1e8 * np.ones((2, 3))
        prob["sigma_y_mat"] = np.array([1e7, 1e7])
        prob["sigma_ult_mat"] = 1e7 * np.ones((2, 3))
        prob["wohler_exp_mat"] = np.array([1e1, 1e1])
        prob["wohler_A_mat"] = np.array([1e1, 1e1])
        prob["rho_mat"] = np.array([1e4, 1e5])
        prob["rho_water"] = 1025.0
        prob["unit_cost_mat"] = np.array([1e1, 2e1])
        prob["outfitting_factor_in"] = 1.1
        prob["material_names"] = ["steel", "slurry"]
        prob["painting_cost_rate"] = 10.0
        prob["labor_cost_rate"] = 2.0

        prob["bulkhead_grid"] = np.array([0.0, 0.1, 0.2, 1.0])
        prob["bulkhead_thickness"] = 1.0 * np.ones(nbulk)

        prob["ring_stiffener_web_thickness"] = 0.2
        prob["ring_stiffener_flange_thickness"] = 0.3
        prob["ring_stiffener_web_height"] = 0.5
        prob["ring_stiffener_flange_width"] = 1.0
        prob["ring_stiffener_spacing"] = 0.2  # non-dimensional ring stiffener spacing

        prob["axial_stiffener_web_thickness"] = 0.2
        prob["axial_stiffener_flange_thickness"] = 0.3
        prob["axial_stiffener_web_height"] = 0.5
        prob["axial_stiffener_flange_width"] = 1.0
        prob["axial_stiffener_spacing"] = 0.25 * np.pi

        prob["ballast_grid"] = np.array([[0.0, 0.1], [0.1, 0.2]])
        prob["ballast_volume"] = np.pi * np.array([10.0, 0.0])

        prob["grid_axial_joints"] = np.array([0.44, 0.55, 0.66])
        prob["joint1"] = np.array([20.0, 10.0, -30.0])
        prob["joint2"] = np.array([25.0, 10.0, 15.0])
        prob["s_ghost1"] = 0.0
        prob["s_ghost2"] = 1.0

        prob.run_model()
        out_list = prob.model.list_outputs(prom_name=True, units=False, out_stream=None)
        for k in out_list:
            if np.all(k[1]["val"] == 0.0) or np.all(k[1]["val"] == NULL):
                name = k[1]["prom_name"]
                if (
                    name.find("Py") > 0
                    or name.find("Pz") > 0
                    or name.find("beta") > 0
                    or name.find("offst") > 0
                    or name.find("tw") >= 0
                ):
                    continue
                print(f"{name} is all zero!")
        self.assertTrue(True)


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestInputs),
        unittest.TestLoader().loadTestsFromTestCase(TestFullDiscretization),
        unittest.TestLoader().loadTestsFromTestCase(TestMemberComponent),
        unittest.TestLoader().loadTestsFromTestCase(TestHydro),
        unittest.TestLoader().loadTestsFromTestCase(TestGlobal2Member),
        unittest.TestLoader().loadTestsFromTestCase(TestGroup),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import unittest

import numpy as np
import numpy.testing as npt

import wisdem.drivetrainse.generator as gen


class TestGenerators(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        self.inputs["machine_rating"] = 5e6
        self.inputs["rated_torque"] = 4.143289e6

        self.inputs["rho_Fe"] = 7700.0
        self.inputs["rho_Fes"] = 7850.0
        self.inputs["rho_Copper"] = 8900.0
        self.inputs["rho_PM"] = 7450.0

        self.inputs["B_r"] = 1.2
        self.inputs["E"] = 2e11
        self.inputs["G"] = 79.3e9
        self.inputs["P_Fe0e"] = 1.0
        self.inputs["P_Fe0h"] = 4.0
        self.inputs["S_N"] = -0.002
        self.inputs["alpha_p"] = 0.5 * np.pi * 0.7
        self.inputs["b_r_tau_r"] = 0.45
        self.inputs["b_ro"] = 0.004
        self.inputs["b_s_tau_s"] = 0.45
        self.inputs["b_so"] = 0.004
        self.inputs["cofi"] = 0.85
        self.inputs["freq"] = 60
        self.inputs["h_i"] = 0.001
        self.inputs["h_sy0"] = 0.0
        self.inputs["h_w"] = 0.005
        self.inputs["k_fes"] = 0.9
        self.inputs["k_fillr"] = 0.7
        self.inputs["k_fills"] = 0.65
        self.inputs["k_s"] = 0.2
        self.discrete_inputs["m"] = 3
        self.inputs["mu_0"] = np.pi * 4e-7
        self.inputs["mu_r"] = 1.06
        self.inputs["p"] = 3.0
        self.inputs["phi"] = np.deg2rad(90)
        self.discrete_inputs["q1"] = 6
        self.discrete_inputs["q2"] = 4
        self.inputs["ratio_mw2pp"] = 0.7
        self.inputs["resist_Cu"] = 1.8e-8 * 1.4
        self.inputs["sigma"] = 40e3
        self.inputs["v"] = 0.3
        self.inputs["y_tau_p"] = 1.0
        self.inputs["y_tau_pr"] = 10.0 / 12
        for k in self.inputs:
            self.inputs[k] = np.array( [self.inputs[k]] )

    def testConstraints(self):
        pass

    def testMofI(self):
        inputs = {}
        outputs = {}
        myobj = gen.MofI()

        inputs["R_out"] = np.array([ 3.0 ])
        inputs["stator_mass"] = np.array([ 60.0 ])
        inputs["rotor_mass"] = np.array([ 40.0 ])
        inputs["generator_mass"] = np.array([ 100.0 ])
        inputs["len_s"] = np.array([ 2.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["generator_I"], np.r_[50.0 * 9, 25.0 * 9 + 100.0 / 3.0, 25.0 * 9 + 100.0 / 3.0])
        npt.assert_almost_equal(outputs["rotor_I"], 0.4 * outputs["generator_I"])
        npt.assert_almost_equal(outputs["stator_I"], 0.6 * outputs["generator_I"])

    def testCost(self):
        inputs = {}
        outputs = {}
        myobj = gen.Cost()

        inputs["C_Cu"] = np.array([ 2.0 ])
        inputs["C_Fe"] = np.array([ 0.5 ])
        inputs["C_Fes"] = np.array([ 4.0 ])
        inputs["C_PM"] = np.array([ 3.0 ])

        inputs["Copper"] = np.array([ 10.0 ])
        inputs["Iron"] = np.array([ 0.0 ])
        inputs["mass_PM"] = np.array([ 0.0 ])
        inputs["Structural_mass"] = np.array([ 0.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["generator_cost"], (1.26 * 2.0 * 10 + 96.2 * 0.064 * 10.0) / 0.619)
 
        inputs["Copper"] = np.array([ 0.0 ])
        inputs["Iron"] = np.array([ 10.0 ])
        inputs["mass_PM"] = np.array([ 0.0 ])
        inputs["Structural_mass"] = np.array([ 0.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["generator_cost"], (1.21 * 0.5 * 10 + 26.9 * 0.064 * 10.0) / 0.684)

        inputs["Copper"] = np.array([ 0.0 ])
        inputs["Iron"] = np.array([ 0.0 ])
        inputs["mass_PM"] = np.array([ 10.0 ])
        inputs["Structural_mass"] = np.array([ 0.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["generator_cost"], (1.0 * 3.0 * 10 + 79.0 * 0.064 * 10.0) / 0.619)

        inputs["Copper"] = np.array([ 0.0 ])
        inputs["Iron"] = np.array([ 0.0 ])
        inputs["mass_PM"] = np.array([ 0.0 ])
        inputs["Structural_mass"] = np.array([ 10.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["generator_cost"], (1.21 * 4.0 * 10 + 15.9 * 0.064 * 10.0) / 0.684)

    def testEff(self):
        inputs = {}
        outputs = {}
        myobj = gen.PowerElectronicsEff()

        inputs["machine_rating"] = np.array([ 2e6 ])
        inputs["shaft_rpm"] = np.arange(10.1)
        inputs["shaft_rpm"][0] = 0.1
        inputs["eandm_efficiency"] = np.ones(inputs["shaft_rpm"].shape)
        myobj.compute(inputs, outputs)
        npt.assert_array_less(outputs["converter_efficiency"], 1.0)
        npt.assert_array_less(0.97, outputs["converter_efficiency"][1:])
        npt.assert_array_less(outputs["transformer_efficiency"], 1.0)
        npt.assert_array_less(0.97, outputs["transformer_efficiency"][1:])
        npt.assert_almost_equal(
            outputs["generator_efficiency"], outputs["transformer_efficiency"] * outputs["converter_efficiency"]
        )

        inputs["machine_rating"] = 20e6
        myobj.compute(inputs, outputs)
        npt.assert_array_less(outputs["converter_efficiency"], 1.0)
        npt.assert_array_less(0.97, outputs["converter_efficiency"][1:])
        npt.assert_array_less(outputs["transformer_efficiency"], 1.0)
        npt.assert_array_less(0.97, outputs["transformer_efficiency"][1:])
        npt.assert_almost_equal(
            outputs["generator_efficiency"], outputs["transformer_efficiency"] * outputs["converter_efficiency"]
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
custom-fix.py

Created by Andrew Ning on 2013-05-23.
Copyright (c) NREL. All rights reserved.
"""

from os import close, remove
from shutil import move
from tempfile import mkstemp

FLAG1 = "[width=5in]{distributedAeroLoads.pdf}"
FLAG2 = "[width=5in]{cp.pdf}"
FLAG3 = "Referring to \\citep{ning2013a-simple-soluti}, this parameter controls"


def fixit(path):
    oldfile = open(path)
    handle, temp_path = mkstemp()
    newfile = open(temp_path, "w")

    for line in oldfile:
        # ---------- Your Custom Replacements Go Here ------------------

        # if FLAG1 in line or FLAG2 in line:
        #     line = line.replace('5in', '4in')

        # if FLAG3 in line:
        #     line = line.replace('citep', 'cite')

        # --------------------------------------------------------------

        newfile.write(line)

    oldfile.close()
    newfile.close()
    close(handle)
    remove(path)
    move(temp_path, path)


if __name__ == "__main__":
    import sys

    path = sys.argv[1]
    fixit(path)



from unittest import TestCase
import pandas as pd
import os
import pytest

from wisdem.landbosse.model import ManagementCost
from wisdem.test.test_landbosse.test_filename_functions import landbosse_test_input_dir

PROJECT_NAME = 'foo'


@pytest.mark.skip(reason="this does not pass")
class TestManagementCost(TestCase):
    def setUp(self):
        """
        This setUp() method executes before each test. It creates an
        instance attribute, self.instance, that refers to a ManagementCost
        instance under test. input_dict is the dictionary of inputs values
        that configures that instance.

        This is just a default dictionary. Methods that test behavior of
        specific variables customize the relevant key/value pairs.

        The key site_facility_building_area_df is set to None here. The
        dataframe is read from the csv in the site_facility tests below.
        """
        # TODO: Refactor the parts of the filenames to the top of the file. Do not hardcode.
        site_facility_building_area_csv = os.path.join(landbosse_test_input_dir(), 'site_facility_building_area.csv')
        site_facility_building_area_df = pd.read_csv(site_facility_building_area_csv)
        self.input_dict = dict()
        self.input_dict['project_value_usd'] = 1e8
        self.input_dict['foundation_cost_usd'] = 1e5
        self.input_dict['construction_time_months'] = 12
        self.input_dict['num_hwy_permits'] = 10
        self.input_dict['num_turbines'] = 10
        self.input_dict['project_size_megawatts'] = 30
        self.input_dict['hub_height_meters'] = 80
        self.input_dict['num_access_roads'] = 2
        self.input_dict['markup_contingency'] = 0.1
        self.input_dict['markup_warranty_management'] = 0.1
        self.input_dict['markup_sales_and_use_tax'] = 0.1
        self.input_dict['markup_overhead'] = 0.1
        self.input_dict['markup_profit_margin'] = 0.1
        self.input_dict['site_facility_building_area_df'] = site_facility_building_area_df
        self.input_dict['construct_duration'] = 9 #months
        self.project_name = 'project_1'
        self.output_dict = dict()
        self.management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)

    def test_ManagementCostModule(self):
        """
        Black box test to check whether module is ran successfully or not
        """
        run_ManagementCost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict,
                                            project_name=self.project_name)
        trial_run = run_ManagementCost.run_module()

        if trial_run[0] == 0:
            print('\n\n================== MODULE EXECUTION SUCCESS =========================\n')
            print(' ManagementCost module ran successfully. See the list of inputs'
                  '\n and outputs below used by the module in its calculations:')
            print('\n=====================================================================\n')

        elif trial_run[0] == 1:
            print(
                '\n\n====================================== MODULE EXECUTION FAILURE ======================================\n')
            print(' > ManagementCost module failed to run successfully. Error detected: ', trial_run[1],
                  '\n > Scroll below to see detailed information about error encountered.'
                  '\n > See the list of inputs below used by the module in its calculations:')
            print(
                '\n========================================================================================================\n')

        print('\nGiven below is the set of inputs fed into ManagementCost module:\n')
        for key, value in self.input_dict.items():
            if isinstance(value, pd.DataFrame):
                print(key, ':\n', value)
            else:
                print(key, ':', value)

        if trial_run[0] == 0:  # Only print outputs if module ran successfully.
            print('\nGiven below is the set of outputs calculated by the RoadsCost module:\n')
            for key, value in self.output_dict.items():
                if isinstance(value, pd.DataFrame):
                    print('\nNow printing DataFrame ->', key, ':\n', value)
                else:
                    print(key, ':', value)

    def test_checks_keys_present(self):
        """
        If not all the keys are present, a ValueError should be raised.
        """
        bad_input_dict = dict()
        bad_input_dict['foundation_cost_usd'] = 1e5
        bad_input_dict['construction_time_months'] = 36
        bad_input_dict['num_hwy_permits'] = 10
        bad_input_dict['markup_constants'] = {}
        bad_input_dict['num_turbines'] = 10
        bad_input_dict['project_size_megawatts'] = 30
        output_dict = dict()
        self.assertRaises(ValueError, ManagementCost, bad_input_dict, output_dict, PROJECT_NAME)

    def test_insurance_cost(self):
        """
        Test the insurance cost calculation
        """
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(5.6e5, self.output_dict['insurance_usd'])

    def test_bonding(self):
        """
        Test the bonding coast calculation
        """
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(1e6, self.output_dict['bonding_usd'])

    def test_project_management_less_than_28_months(self):
        """
        Tests project management cost calculation less than 28 months.
        Ensure that the instance is set to 12 months for this calculation.
        """
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(2011724, round(self.output_dict['project_management_usd']))

    def test_project_management_greater_than_28_months(self):
        self.input_dict['construction_time_months'] = 36
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(2011724, round(self.output_dict['project_management_usd']))

    def test_markup_contingency_costs(self):
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(5e7, self.output_dict['markup_contingency_usd'])

    def test_engineering_under_200(self):
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(1104760.0, self.output_dict['engineering_usd'])

    def test_engineering_over_or_equal_200(self):
        self.input_dict['project_size_megawatts'] = 300
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(1451535, self.output_dict['engineering_usd'])

    def test_site_facility_1(self):
        self.input_dict['project_size_megawatts'] = 100
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(928000.0, self.output_dict['site_facility_usd'])

    def test_total_management_cost(self):
        management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=PROJECT_NAME)
        management_cost.run_module()
        self.assertEqual(55802284, round(self.output_dict['total_management_cost']))



import os
import textwrap
import subprocess

import numpy as np


class DakotaOptimizer:
    def __init__(self, template_dir="template_dir/"):
        self.template_dir = template_dir
        self.setup_directories(template_dir)

    def optimize(self, desvars, desired_outputs, bounds, model_string, output_scalers, options=None):
        """
        Helper method that calls all of the other methods to do the full
        top-to-bottom optimization process.
        """

        desvar_labels, desvar_shapes = self.create_input_file(
            self.template_dir, desvars, desired_outputs, bounds, options
        )
        self.create_input_yaml(self.template_dir, desvar_labels)
        self.create_driver_file(self.template_dir, model_string, desvar_shapes, desired_outputs, output_scalers)
        self.run_dakota()
        results = self.postprocess()
        return results

    def setup_directories(self, template_dir):
        """
        Create the template_dir, remove previous files.
        """
        if not os.path.exists(template_dir):
            os.makedirs(template_dir)

        subprocess.call("rm *.in", shell=True)
        subprocess.call("rm *.out", shell=True)
        subprocess.call("rm *.rst", shell=True)
        subprocess.call("rm *.dat", shell=True)
        subprocess.call("rm -rf run_history", shell=True)

    def create_input_file(self, template_dir, desvars, outputs, bounds, options):
        """
        Create the Dakota input file based on the user-defined problem formulation.

        Basically, we need the number and size of design variables, their bounds,
        and the functions of interest. Here is also where you set the optimization
        method and termination criteria.
        """
        flattened_bounds = []

        for key, value in bounds.items():
            if isinstance(value, (float, list)):
                value = np.array(value)
            flattened_value = np.squeeze(value.flatten()).reshape(-1, 2)
            flattened_bounds.extend(flattened_value)

        flattened_bounds = np.array(flattened_bounds)

        desvar_labels = []
        for key, value in desvars.items():
            if isinstance(value, (float, list)):
                value = np.array(value)
            flattened_value = np.squeeze(value.flatten())
            key = key.replace(".", "_")
            for i in range(flattened_value.size):
                desvar_labels.append(f"{key}_{i}")

        desvar_shapes = {}
        total_size = 0

        for key, value in desvars.items():
            if isinstance(value, (float, list)):
                value = np.array(value)

            desvar_shapes[key] = value.shape
            total_size += value.size

        if "cobyla" in options["method"]:
            opt_options = {
                "max_iterations": 100,
                "max_function_evaluations": 200,
                "initial_delta": 0.5,
            }
            opt_options.update(options)

        elif "efficient_global" in options["method"]:
            opt_options = {
                "seed": 123,
            }
            opt_options.update(options)

        method = opt_options.pop("method")

        options_string_list = []
        for option, val in opt_options.items():
            line = f"    {option} = {val}"
            options_string_list.append(line)
        options_string_lines = """
    {}
        """.format(
            "\n".join(options_string_list)
        )

        if len(outputs) > 1:
            constraints_string = f"  nonlinear_equality_constraints {len(outputs)-1}\n"
        else:
            constraints_string = ""

        #### Flatten the DVs here and make names for each and append the names with the number according to the size of the DVs

        # Terrible string-list manipulation to get the DVs and outputs formatted correctly
        input_file = (
            textwrap.dedent(
                """\
        # Dakota input file
        environment
          tabular_data
            tabular_data_file "dakota_data.dat"

        method
        """
            )
            + f"  {method}"
            + options_string_lines
            + textwrap.dedent(
                """
        variables
        """
            )
            + f"  continuous_design {len(flattened_bounds)}\n"
            + f"  lower_bounds "
            + " ".join([str(i) for i in flattened_bounds[:, 0]])
            + "\n"
            + f"  upper_bounds "
            + " ".join([str(i) for i in flattened_bounds[:, 1]])
            + textwrap.dedent(
                """

        interface
          fork
            asynchronous
            evaluation_concurrency 1
            parameters_file "params.in"
            results_file "results.out"
        """
            )
            + f'    copy_files "{template_dir}*"'
            + textwrap.dedent(
                """
            analysis_driver "python openmdao_driver.py"

            work_directory
              named "run_history/run"
              directory_tag
              directory_save
              file_save

        responses
          objective_functions 1
        """
            )
            + constraints_string
            + "  descriptors "
            + " ".join(['"' + key + '"' for key in outputs])
            + """
      no_gradients
      no_hessians
        """
        )

        with open("dakota_input.in", "w") as text_file:
            text_file.write(input_file)

        return desvar_labels, desvar_shapes

    def create_input_yaml(self, template_dir, desvar_labels):
        """
        Create the input yaml file which is just a file with a dictionary format
        so that Dakota knows which design variables to populate in the problem.
        """
        # Populate input.yml
        input_lines = [f"cdv_{i+1}: {{cdv_{i+1}}}" for i in range(len(desvar_labels))]
        with open(template_dir + "input_template.yml", "w") as f:
            for line in input_lines:
                f.write(line + "\n")

    def create_driver_file(self, template_dir, model_string, desvar_shapes, desired_outputs, output_scalers):
        """
        Create the Python file that actually runs the model analysis.

        Once created, this file will take in the design variables as set by Dakota
        and output all functions of interest. This function that creates that file
        needs the model definition (import statement), as well as the desvar sizes,
        the list of desired outputs, and the output scalers.
        """
        desvar_shapes_lines = []
        for key in desvar_shapes:
            desvar_shapes_lines.append(f'desvar_shapes["{key}"] = {desvar_shapes[key]}')
        desvar_shapes_lines = """
{}
        """.format(
            "\n".join(desvar_shapes_lines)
        )

        desired_outputs_string = "desired_outputs = [" + " ".join(['"' + key + '"' for key in desired_outputs]) + "]"

        write_outputs_string = []
        for i, key in enumerate(desired_outputs):
            string = f'f.write(str(float(outputs["{key}"]) * {output_scalers[i]}) + "\\n")'
            write_outputs_string.append(string)

        write_outputs_string = """
    {}
        """.format(
            "\n    ".join(write_outputs_string)
        )

        # Create openmdao_driver.py
        driver_file = (
            textwrap.dedent(
                """\
        # Import  modules
        import sys
        from subprocess import call
        import numpy as np
        from yaml import safe_load


        #########################################
        #                                       #
        #    Step 1: Use Dakota created         #
        #    input files to prepare for         #
        #    model run.                         #
        #                                       #
        #########################################
        input_template = "input_template.yml"
        inputs = "inputs.yml"
        call(["dprepro", sys.argv[1], input_template, inputs])
        call(['rm', input_template])

        #########################################
        #                                       #
        #    Step 2: Run Model                  #
        #                                       #
        #########################################
        # Load parameters from the yaml formatted input.
        with open(inputs, "r") as f:
            desvars = safe_load(f)

        desvars_list = []
        for key in desvars:
            desvars_list.append(float(desvars[key]))
        flattened_desvars = np.array(desvars_list)

        desvar_shapes = {}
        """
            )
            + desvar_shapes_lines
            + textwrap.dedent(
                """
        size_counter = 0
        desvars = {}
        for key, shape in desvar_shapes.items():
            size = int(np.prod(shape))
            desvars[key] = flattened_desvars[
                size_counter : size_counter + size
            ].reshape(shape)
            size_counter += size

        print()
        print('Design variables:')
        print(desvars)
        """
            )
            + model_string
            + "\n"
            + textwrap.dedent(
                """\
        model_instance = model(desvars)
        outputs = model_instance.compute(desvars)
        #########################################
        #                                       #
        #    Step 3: Write Output in format     #
        #    Dakota expects                     #
        #                                       #
        #########################################
        """
            )
            + desired_outputs_string
            + "\n"
            + textwrap.dedent(
                """\
        print('Outputs:')
        print(outputs)
        # Write it to the expected file.
        with open(sys.argv[2], "w") as f:"""
            )
            + write_outputs_string
        )

        with open(template_dir + "openmdao_driver.py", "w") as text_file:
            text_file.write(driver_file)

    def run_dakota(self):
        """
        Actually run Dakota from the command line.
        """
        subprocess.call("dakota -i dakota_input.in -o dakota_output.out -write_restart dakota_restart.rst", shell=True)
        subprocess.call("dakota_restart_util to_tabular dakota_restart.rst dakota_data.dat", shell=True)

    def postprocess(self):
        """
        Postprocess the dakota_data.dat output file and create a dictionary
        with run results.
        """

        results = {}
        key_list = []
        with open("dakota_data.dat") as f:
            for i, line in enumerate(f):
                # Grab headers from first line
                if i == 0:
                    for key in line.split():
                        results[key] = []
                        key_list.append(key)
                if i > 0:
                    split_line = line.split()
                    for j, key in enumerate(key_list):
                        try:
                            # If it's a float, cast it as such
                            column_entry = float(split_line[j])
                        except ValueError:
                            # If it's an actual string, keep it as a string
                            column_entry = split_line[j]
                        results[key].append(column_entry)

        return results



import unittest

import openmdao.api as om

import wisdem.nrelcsm.nrel_csm_mass_2015 as nct2015


class TestAll(unittest.TestCase):
    def testMass(self):
        # simple test of module
        prob = om.Problem(reports=False)
        prob.model = nct2015.nrel_csm_mass_2015()
        prob.setup()

        prob["rotor_diameter"] = 126.0
        prob["turbine_class"] = 1
        prob["blade_has_carbon"] = False
        prob["blade_number"] = 3
        prob["machine_rating"] = 5000.0
        prob["tower_length"] = 90.0
        prob["main_bearing_number"] = 2
        prob["crane"] = True
        prob["max_tip_speed"] = 80.0
        prob["max_efficiency"] = 0.90

        prob.run_model()

        self.assertAlmostEqual(float(prob["blade_mass"][0]), 18590.66820649, 2)
        self.assertAlmostEqual(float(prob["hub_mass"][0]), 44078.53687493, 2)
        self.assertAlmostEqual(float(prob["pitch_system_mass"][0]), 10798.90594644, 2)
        self.assertAlmostEqual(float(prob["spinner_mass"][0]), 973.0, 2)
        self.assertAlmostEqual(float(prob["lss_mass"][0]), 22820.27928238, 2)
        self.assertAlmostEqual(float(prob["main_bearing_mass"][0]), 2245.41649102, 2)
        self.assertAlmostEqual(float(prob["rated_rpm"][0]), 12.1260909, 2)
        self.assertAlmostEqual(float(prob["rotor_torque"][0]), 4375.0, 2)
        self.assertAlmostEqual(float(prob["gearbox_mass"][0]), 21875.0, 2)
        self.assertAlmostEqual(float(prob["hss_mass"][0]), 994.7, 2)
        self.assertAlmostEqual(float(prob["generator_mass"][0]), 14900.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_mass"][0]), 41765.26095285, 2)
        self.assertAlmostEqual(float(prob["yaw_mass"][0]), 12329.96247921, 2)
        self.assertAlmostEqual(float(prob["hvac_mass"][0]), 400.0, 2)
        self.assertAlmostEqual(float(prob["cover_mass"][0]), 6836.69, 2)
        self.assertAlmostEqual(float(prob["platforms_mass"][0]), 8220.65761911, 2)
        self.assertAlmostEqual(float(prob["transformer_mass"][0]), 11485.0, 2)
        self.assertAlmostEqual(float(prob["tower_mass"][0]), 182336.48057717, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass"][0]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["rotor_mass"][0]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass"][0]),  151455.88331558352, 2)
        self.assertAlmostEqual(float(prob["turbine_mass"][0]), 445414.81133358914, 2)

    def testMassAndCost(self):
        # simple test of module
        prob = om.Problem(reports=False)
        prob.model = nct2015.nrel_csm_2015()
        prob.setup()

        prob["rotor_diameter"] = 126.0
        prob["turbine_class"] = 1
        prob["blade_has_carbon"] = False
        prob["blade_number"] = 3
        prob["machine_rating"] = 5000.0
        prob["tower_length"] = 90.0
        prob["main_bearing_number"] = 2
        prob["crane"] = True
        prob["max_tip_speed"] = 80.0
        prob["max_efficiency"] = 0.90

        prob.run_model()

        self.assertAlmostEqual(float(prob["blade_mass"][0]), 18590.66820649, 2)
        self.assertAlmostEqual(float(prob["hub_mass"][0]), 44078.53687493, 2)
        self.assertAlmostEqual(float(prob["pitch_system_mass"][0]), 10798.90594644, 2)
        self.assertAlmostEqual(float(prob["spinner_mass"][0]), 973.0, 2)
        self.assertAlmostEqual(float(prob["lss_mass"][0]), 22820.27928238, 2)
        self.assertAlmostEqual(float(prob["main_bearing_mass"][0]), 2245.41649102, 2)
        self.assertAlmostEqual(float(prob["rated_rpm"][0]), 12.1260909, 2)
        self.assertAlmostEqual(float(prob["rotor_torque"][0]), 4375.0, 2)
        self.assertAlmostEqual(float(prob["gearbox_mass"][0]), 21875.0, 2)
        self.assertAlmostEqual(float(prob["hss_mass"][0]), 994.7, 2)
        self.assertAlmostEqual(float(prob["generator_mass"][0]), 14900.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_mass"][0]), 41765.26095285, 2)
        self.assertAlmostEqual(float(prob["yaw_mass"][0]), 12329.96247921, 2)
        self.assertAlmostEqual(float(prob["hvac_mass"][0]), 400.0, 2)
        self.assertAlmostEqual(float(prob["cover_mass"][0]), 6836.69, 2)
        self.assertAlmostEqual(float(prob["platforms_mass"][0]), 8220.65761911, 2)
        self.assertAlmostEqual(float(prob["transformer_mass"][0]), 11485.0, 2)
        self.assertAlmostEqual(float(prob["tower_mass"][0]), 182336.48057717, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass"][0]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["rotor_mass"][0]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass"][0]),  151455.88331558352, 2)
        self.assertAlmostEqual(float(prob["turbine_mass"][0]), 445414.81133358914, 2)

        self.assertAlmostEqual(float(prob["blade_cost"][0]), 271423.75581475, 2)
        self.assertAlmostEqual(float(prob["hub_cost"][0]), 171906.29381221, 2)
        self.assertAlmostEqual(float(prob["pitch_system_cost"][0]), 238655.82141628, 2)
        self.assertAlmostEqual(float(prob["spinner_cost"][0]), 10800.3, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass_tcc"][0]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["hub_system_cost"][0]), 421362.41522849, 2)
        self.assertAlmostEqual(float(prob["rotor_cost"][0]), 1235633.68267274, 2)
        self.assertAlmostEqual(float(prob["rotor_mass_tcc"][0]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["lss_cost"][0]), 271561.32346034, 2)
        self.assertAlmostEqual(float(prob["main_bearing_cost"][0]), 10104.37420958, 2)
        self.assertAlmostEqual(float(prob["gearbox_cost"][0]), 218750.0 , 2)
        self.assertAlmostEqual(float(prob["hss_cost"][0]), 6763.96, 2)
        self.assertAlmostEqual(float(prob["generator_cost"][0]), 184760.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_cost"][0]), 121119.25676326, 2)
        self.assertAlmostEqual(float(prob["yaw_system_cost"][0]), 102338.68857747, 2)
        self.assertAlmostEqual(float(prob["hvac_cost"][0]), 49600.0, 2)
        self.assertAlmostEqual(float(prob["controls_cost"][0]), 105750.0, 2)
        self.assertAlmostEqual(float(prob["converter_cost"][0]), 0.0, 2)
        self.assertAlmostEqual(float(prob["elec_cost"][0]), 209250.0, 2)
        self.assertAlmostEqual(float(prob["cover_cost"][0]), 38969.133, 2)
        self.assertAlmostEqual(float(prob["platforms_cost"][0]), 101273.24528671, 2)
        self.assertAlmostEqual(float(prob["transformer_cost"][0]), 215918.0, 2)
        self.assertAlmostEqual(float(prob["nacelle_cost"][0]), 1665612.9280069391, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass_tcc"][0]), 151455.88331558352 , 2)
        self.assertAlmostEqual(float(prob["tower_parts_cost"][0]), 528775.7936738, 2)
        self.assertAlmostEqual(float(prob["tower_cost"][0]), 528775.7936738, 2)
        self.assertAlmostEqual(float(prob["turbine_mass_tcc"][0]), 445414.81133358914, 2)
        self.assertAlmostEqual(float(prob["turbine_cost"][0]), 3430022.404353479, 2)
        self.assertAlmostEqual(float(prob["turbine_cost_kW"][0]), 686.0044808706958, 2)


if __name__ == "__main__":
    unittest.main()
